/*
 *
 * These 4 examples demonstrate general purpose computing by GPGPU in GLSL.
 *
 * Command to compile: g++ gpgpu_computing_glsl.cpp -lfreeglut -lglew32 -lopengl32 -lglu32 -lwinmm -lgdi32 -DEXAMPLE1
 * to compile example1 code. Similarly, -DEXAMPLE2 for EXAMPLE2 code, etc
 *
 * Alternatively, use #define EXAMPLE1 for compile the EXAMPLE1 code.
 *
 * These examples ware adopted from the following source.
 *
 * (c) 2006 Dominik G�ddeke, University of Dortmund
 * dominik.goeddeke@math.uni-dortmund.de
 * For details, please refer to the tutorial available at
 * www.mathematik.uni-dortmund.de/~goeddeke/gpgpu/tutorial3.html
 */

//#define EXAMPLE1
//#define EXAMPLE2
//#define EXAMPLE3
//#define EXAMPLE4

#include <stdio.h>
#include <stdlib.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <time.h>
#include <string>
#include <assert.h>

// common functions
GLuint setupShader(const char **vertSrc, const char **fragSrc);
void CHECK_FRAMEBUFFER_STATUS();
void copy(float* in, float* out, int N);
void copy(float* in, void* out, int N);
void copy(void* in, float* out, int N);

#ifdef EXAMPLE1

// this example demonstrates data transfer between CPU and GPU sides

int main(int argc, char **argv) {
	// declare texture size, the actual data will be a vector
	// of size texSize*texSize*4
	int texSize = 2;

	// create test data
	float* data = (float*) malloc(4*texSize*texSize*sizeof(float));
	float* result = (float*) malloc(4*texSize*texSize*sizeof(float));

	for (int i=0; i<texSize*texSize*4; i++) {
		data[i] = i+1.0;
		result[i] = 0.0;
	}

	// set up glut to get valid GL context and
	// get extension entry points
	glutInit(&argc, argv);
	glutCreateWindow("Example 1");
	//glutShowWindow();
	glutHideWindow();
	//glutIconifyWindow();

	glewInit();
	// viewport transform for 1:1 pixel=texel=data mapping
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0,texSize,0.0,texSize);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0,0,texSize,texSize);

	// create FBO and bind it (that is, use offscreen render target)
	GLuint fb;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fb);

	// create texture
	GLuint tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D,tex);

	// set texture parameters
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// define texture with GL_RGBA format and floating point type
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, texSize,texSize,0,GL_RGBA,GL_FLOAT,0);

	// attach texture
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,tex,0);

	// transfer data to texture
	glTexSubImage2D(GL_TEXTURE_2D,0,0,0,texSize,texSize, GL_RGBA,GL_FLOAT,data);

	// and read back
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glReadPixels(0, 0, texSize, texSize,GL_RGBA,GL_FLOAT,result);

	// print out results
	printf("Data before and after:\n");
	for (int i=0; i<texSize*texSize*4; i++) {
		printf("%d: %0.1f, %0.1f\n", i, data[i], result[i]);
	}

	// clean up
	free(data);
	free(result);
	glDeleteFramebuffersEXT (1,&fb);
	glDeleteTextures (1,&tex);
	return 0;
}

#endif

#ifdef EXAMPLE2

// this example demonstrate GPU computing on one texture

// fragment shader as kernel program in glsl
static const char *fragSource = {
"uniform sampler2D texUnit;"
"void main(void)"
"{"
"   vec4 texVal  = texture2D(texUnit, gl_TexCoord[0].xy);"
"   gl_FragColor.x = texVal[0]*2;"
"   gl_FragColor.y = texVal[1]*3;"
"   gl_FragColor.z = texVal[2]/2;"
"   gl_FragColor.a = sqrt(texVal[3]);"
//"gl_FragColor = texVal;"
//"gl_FragColor = sqrt(texVal);"
"}"
};

int main(int argc, char **argv) {
	// define texture size
	int texSize = 4;
	// declare float array of length texSize*texSize*4 for input data
	float data[64] = { 24, 2, 3, 4, 5, 64, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
			17, 18, 19, 20, 21, 22, 23, 1, 25, 53, 27, 28, 29, 30, 31, 32, 33,
			34, 35, 36, 37, 38, 39, 40, 41, 42, 46, 44, 45, 43, 47, 48, 49, 50,
			51, 52, 26, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 6 };

	//for reading back
	float* result = (float*) malloc(4 * texSize * texSize * sizeof(float));
	for (int i = 0; i < texSize * texSize * 4; i++) {
		result[i] = 0.0;
	}

	glutInit(&argc, argv);
	glutCreateWindow("Example 2");
	glewInit();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, texSize, 0.0, texSize);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0, 0, texSize, texSize);

	// create FBO and bind it (that is, use offscreen render target)
	GLuint fb;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

	// create texture by data
	GLuint tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize, texSize, 0, GL_RGBA, GL_FLOAT, data);

	// create destination texture
	GLuint desttex;
	glGenTextures(1, &desttex);
	glBindTexture(GL_TEXTURE_2D, desttex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	// define texture with floating point format
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize, texSize, 0, GL_RGBA, GL_FLOAT, NULL);
	//CHECK_FRAMEBUFFER_STATUS();

	// attach texture as frame buffer
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, desttex, 0);
	//CHECK_FRAMEBUFFER_STATUS();

	// create fragment shader on texture data computing, pixel by pixel
	GLuint shader;
	shader = setupShader(NULL, &fragSource);
	glUseProgram(shader);

	GLint loc = glGetUniformLocation(shader, "texUnit");
	glUniform1i(loc, 0);          // set texture unit 0
	glActiveTexture(GL_TEXTURE0); // activate Texture unit 0
	glBindTexture(GL_TEXTURE_2D, tex);

// draw
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(0, 0);
	glTexCoord2f(1, 0);
	glVertex2f(texSize, 0);
	glTexCoord2f(1, 1);
	glVertex2f(texSize, texSize);
	glTexCoord2f(0, 1);
	glVertex2f(0, texSize);
	glEnd();
	glFlush();

	// reading back
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glReadPixels(0, 0, texSize, texSize, GL_RGBA, GL_FLOAT, result);

	// print out results
	printf("Index: data, result\n");
	for (int i = 0; i < texSize * texSize * 4; i++) {
		printf("%d: %0.1f, %0.1f\n", i, data[i], result[i]);
	}

	// clean up
	free(result);
	glDeleteFramebuffersEXT(1, &fb);
	glDeleteTextures(1, &tex);
	return 0;
}

#endif

#ifdef EXAMPLE3

// this example shows GPU computing on two textures

// fragment shader as kernel program in GLSL
static const char *fragSource =
{
	"uniform sampler2D texUnit;"
	"uniform sampler2D texUnit1;"
	"uniform float alpha;"
	"uniform float selection;"
	//"in vec2 texCoord;"
	"void main(void)"
	"{"
	"   vec4 texVal  = texture2D(texUnit, gl_TexCoord[0].xy);"
	"   vec4 texVal1  = texture2D(texUnit1, gl_TexCoord[0].xy);"
	"   if (selection <= 1.5) {"
	"gl_FragColor = texVal+texVal1;"
//"   gl_FragColor.x = texVal[0];"
//"   gl_FragColor.y = texVal[1];"
//"   gl_FragColor.z = texVal[0] + texVal[1];"
//"   gl_FragColor.a = selection;"
	"   } else {"
	"   gl_FragColor.x = texVal[0]+texVal[1]+texVal[2]+texVal[3]+texVal1[0]+texVal1[1]+texVal1[2]+texVal1[3];"
	"   gl_FragColor.y =  10;"
	"   gl_FragColor.z = alpha;"
	"   gl_FragColor.a = selection;"
	"   }"
	"}"
};

int main(int argc, char **argv)
{
	// declare texture size, the actual data will be a vector
	// of size texSize*texSize*4
	int texSize = 2;
	// create test data
	float* result = (float*)malloc(4*texSize*texSize*sizeof(float));
	float* data = (float*) malloc(4*texSize*texSize*sizeof(float));
	float* data1 = (float*) malloc(4*texSize*texSize*sizeof(float));

	for (int i=0; i<texSize*texSize; i++) {
		data[i*4] = i*4 + 1.0;
		data[i*4+1] = i*4 + 2.0;
		data[i*4+2] = i*4 +3.0;
		data[i*4+3] = i*4 +4.0;
	}

	for (int i=0; i<texSize*texSize; i++) {
		data1[i*4] = i*4 +5.0;
		data1[i*4+1] = i*4 + 6.0;
		data1[i*4+2] = i*4 +7.0;
		data1[i*4+3] = i*4 +8.0;
	}

	for (int i=0; i<texSize*texSize*4; i++) {
		result[i] = 0.0;
	}

	glutInit (&argc, argv);
	glutCreateWindow("Example 3");
	glewInit();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0,texSize,0.0,texSize);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0,0,texSize,texSize);

	// create FBO and bind it (that is, use offscreen render target)
	GLuint fb;
	glGenFramebuffersEXT(1,&fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fb);

	// create data textures
	GLuint tex[2];
	glGenTextures (2, tex);

	glBindTexture(GL_TEXTURE_2D,tex[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, texSize,texSize,0,GL_RGBA,GL_FLOAT, data);

	glBindTexture(GL_TEXTURE_2D,tex[1]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, texSize,texSize,0,GL_RGBA,GL_FLOAT, data1);

	// create destination texture
	GLuint desttex;
	glGenTextures (1, &desttex);
	glBindTexture(GL_TEXTURE_2D,desttex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB,texSize,texSize,0,GL_RGBA,GL_FLOAT, NULL);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,desttex,0);
	//CHECK_FRAMEBUFFER_STATUS();

	GLuint shader;
	shader = setupShader(NULL, &fragSource);
	glUseProgram(shader);
//glUseProgram(0);


	GLint loc = glGetUniformLocation(shader, "texUnit");
	glUniform1i(loc, 0); // texture unit 0
	glActiveTexture(GL_TEXTURE0);// Texture unit 0
	glBindTexture(GL_TEXTURE_2D,tex[0]);

	GLint loc1 = glGetUniformLocation(shader, "texUnit1");
	glUniform1i(loc1, 1);// texture unit 1
	glActiveTexture(GL_TEXTURE1);// Texture unit 1
	glBindTexture(GL_TEXTURE_2D,tex[1]);


	GLint alphapar = glGetUniformLocation(shader, "alpha");
	glUniform1f(alphapar, 0.5);// texture unit 0

	GLint selectionpar = glGetUniformLocation(shader, "selection");

	// draw1
	glUniform1f(selectionpar, 1); // set the case selection <= 1.5
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(0, 0);
	glTexCoord2f(1, 0);
	glVertex2f(texSize, 0);
	glTexCoord2f(1, 1);
	glVertex2f(texSize, texSize);
	glTexCoord2f(0, 1);
	glVertex2f(0, texSize);
	glEnd();
	glFlush();

	// and read back
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glReadPixels(0, 0, texSize, texSize,GL_RGBA,GL_FLOAT,result);

	// print out results
	printf("Index: data, data1, result (selection 1)\n");
	for (int i=0; i<texSize*texSize*4; i++) {
		printf("%d: %0.1f, %0.1f, %0.1f\n", i, data[i], data1[i], result[i]);
	}

	// draw2

	glUniform1f(selectionpar, 2); // set the case selection > 1.5

	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(0, 0);
	glTexCoord2f(1, 0);
	glVertex2f(texSize, 0);
	glTexCoord2f(1, 1);
	glVertex2f(texSize, texSize);
	glTexCoord2f(0, 1);
	glVertex2f(0, texSize);
	glEnd();
	glFlush();

	// Read back
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glReadPixels(0, 0, texSize, texSize,GL_RGBA,GL_FLOAT,result);

	// print out results
	printf("Index: Data, Data1, Result (selection 2)\n");
	for (int i=0; i<texSize*texSize*4; i++) {
		printf("%d: %0.1f, %0.1f, %0.1f\n", i, data[i], data1[i], result[i]);
	}

	// clean up
	free(result);
	free(data);
	free(data1);
	glDeleteFramebuffersEXT (1,&fb);
	glDeleteTextures(1,tex);

	return 0;
}

#endif

#ifdef EXAMPLE4

// this example demonstrates pingpong style of computing with multiple iterations and multiple textures.

// array size N is texsize*texsize
static const int texSize = 256; //1025
static const int N = texSize * texSize;

// hard-coded for shader + Cg, do not change this
static const int numArraysPerChunk = 9;

// input data size, default setting results in approx 1.6 GB of phys mem.
static const int numChunks = 20;

// number of iterations per chunk to emulate kernel workload
static const int numKernelSteps = 1;

// number of iterations the test is repeated, to reduce timing noise.
static const int numIterations = 10;

#define DEBUGDEBUG

static const char* kernelSource = {
"uniform sampler2D array0;"
"uniform sampler2D array1;"
"uniform sampler2D array2;"
"uniform sampler2D array3;"
"uniform sampler2D array4;"
"uniform sampler2D array5;"
"uniform sampler2D array6;"
"uniform sampler2D array7;"
"uniform sampler2D array8;"
"uniform sampler2D accum;"
"void main(void)"
"{"
//"vec4 texVal0  = texture2D(array0, gl_TexCoord[0].xy);"
//"vec4 texVal1  = texture2D(array1, gl_TexCoord[0].xy);"
//"vec4 texVal2  = texture2D(array2, gl_TexCoord[0].xy);"
//"vec4 texVal3  = texture2D(array3, gl_TexCoord[0].xy);"
//"vec4 texVal4  = texture2D(array4, gl_TexCoord[0].xy);"
//"vec4 texVal5  = texture2D(array5, gl_TexCoord[0].xy);"
//"vec4 texVal6  = texture2D(array6, gl_TexCoord[0].xy);"
//"vec4 texVal7  = texture2D(array7, gl_TexCoord[0].xy);"
//"vec4 texVal8  = texture2D(array8, gl_TexCoord[0].xy);"
//"vec4 accumVal  = texture2D(accum, gl_TexCoord[0].xy);"
//"gl_FragColor.x = 100; "
//"gl_FragColor.y = accumVal[0] + texVal1[0] + texVal2[0] ; "
//"gl_FragColor = accumVal + texVal1 + texVal2;"
//"gl_FragColor = accumVal + texVal0+ texVal1+ texVal2+ texVal3+ texVal4+ texVal5+ texVal6+ texVal7 + texVal8; "
"gl_FragColor = "
"texture2D(accum, gl_TexCoord[0].xy) +  "
"texture2D(array0, gl_TexCoord[0].xy) +   "
"texture2D(array1, gl_TexCoord[0].xy) +   "
"texture2D(array2, gl_TexCoord[0].xy) +   "
"texture2D(array3, gl_TexCoord[0].xy) +   "
"texture2D(array4, gl_TexCoord[0].xy) +   "
"texture2D(array5, gl_TexCoord[0].xy) +   "
"texture2D(array6, gl_TexCoord[0].xy) +   "
"texture2D(array7, gl_TexCoord[0].xy) +   "
"texture2D(array8, gl_TexCoord[0].xy);  "
"};"
};


#define DEBUGDEBUG

int main(int argc, char **argv) {
	assert (numArraysPerChunk == 9);

	glutInit(&argc, argv);
	glutCreateWindow("Example 4");
	glutHideWindow();
	glewInit();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, texSize, 0.0, texSize);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0, 0, texSize, texSize);

	GLuint fb;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);


	printf("---------------------------------------------\n");
	printf("Creating input data, if the following amount \n");

	double memrequirements = (numChunks * numArraysPerChunk + numChunks) * N * sizeof(float);

	// single precision textures
	memrequirements += (numArraysPerChunk + 2) * N * sizeof(float);

	// convert to megabytes
	memrequirements /= (1024.0 * 1024.0);

	printf("Allocating %.2f MB.\n", memrequirements);

	// input arrays: arrays in chunk one contain values from 1..9,
	// in chunk two from 10..19, and so on
	printf("The computing is to add the following values for %d times\n", numKernelSteps);
	printf("The same computing is done on all texSize*texSize*4=%d positions\n", texSize*texSize*4);
	float*** chunks = new float**[numChunks];
	for (int chunk = 0; chunk < numChunks; chunk++) {
		chunks[chunk] = new float*[numArraysPerChunk];
		for (int array = 0; array < numArraysPerChunk; array++) {
			chunks[chunk][array] = new float[N * 4];
			for (int i = 0; i < N; i++) {
				chunks[chunk][array][4 * i] =     chunk * 10.0 + array + 1;
				chunks[chunk][array][4 * i + 1] = chunk * 10.0 + array + 1;
				chunks[chunk][array][4 * i + 2] = chunk * 10.0 + array + 1;
				chunks[chunk][array][4 * i + 3] = chunk * 10.0 + array + 1;
			}
			printf("%.1f ", chunks[chunk][array][0]);
		}
		printf("\n");
	}


	float* tmpfloat = new float[N * 4];

	// create two textures for pingpong intermediate memory, and initialization
	GLuint pingtex[2];
	glGenTextures(2, pingtex);
	float* data = new float[N * 4];
	for (int i = 0; i < N*4; i++) {
		data[i] = 0;
	}
	copy(data, tmpfloat, 4 * N);
	for (int i = 0; i < 2; i++) {
		glBindTexture(GL_TEXTURE_2D, pingtex[i]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize, texSize, 0, GL_RGBA, GL_FLOAT, NULL);
	}

	for (int i = 0; i < 2; i++) {
		glBindTexture(GL_TEXTURE_2D, pingtex[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize, texSize, 0, GL_RGBA, GL_FLOAT, tmpfloat);
	}

	// create numArraysPerChunk input textures, set parameters
	GLuint inputTextures[numArraysPerChunk];
	glGenTextures(numArraysPerChunk, inputTextures);
	for (int array = 0; array < numArraysPerChunk; array++) {
		glBindTexture(GL_TEXTURE_2D, inputTextures[array]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize, texSize, 0, GL_RGBA, GL_FLOAT, NULL);
	}


	//
	// set up GLSL runtime
	// create kernel program, load, compile and link it
	//
	GLuint glsl_shader = setupShader(NULL, &kernelSource);
	glUseProgram(glsl_shader);


	//
	// assemble input parameters to kernel
	//
	GLenum texunits[] = {GL_TEXTURE0, GL_TEXTURE1, GL_TEXTURE2, GL_TEXTURE3, GL_TEXTURE4, GL_TEXTURE5, GL_TEXTURE6, GL_TEXTURE7, GL_TEXTURE8};
	GLint params[9];
	params[0] = glGetUniformLocation(glsl_shader, "array0");
	params[1] = glGetUniformLocation(glsl_shader, "array1");
	params[2] = glGetUniformLocation(glsl_shader, "array2");
	params[3] = glGetUniformLocation(glsl_shader, "array3");
	params[4] = glGetUniformLocation(glsl_shader, "array4");
	params[5] = glGetUniformLocation(glsl_shader, "array5");
	params[6] = glGetUniformLocation(glsl_shader, "array6");
	params[7] = glGetUniformLocation(glsl_shader, "array7");
	params[8] = glGetUniformLocation(glsl_shader, "array8");
	GLint accumParam = glGetUniformLocation(glsl_shader, "accum");

	//
	// attach ping pong textures to FBO,
	// init pingpong management variables
	//
	GLuint readTex = 0;
	GLuint writeTex = 1;
	GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[readTex], GL_TEXTURE_2D, pingtex[readTex], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[writeTex], GL_TEXTURE_2D, pingtex[writeTex], 0);


	//
	// perform computation by looping over all chunks of input data,
	// streaming them in as we proceed.
	//
	printf("Starting calculation\n");
	glFinish();
	double start_compute = clock();


	for (int iter = 0; iter < numIterations; iter++) {
		for (int chunk = 0; chunk < numChunks; chunk++) {			//
			// populate input textures for current chunk and transfer data to driver-controlled memory.
			for (int array = 0; array < numArraysPerChunk; array++) {
				copy(chunks[chunk][array], tmpfloat, 4 * N);
				glBindTexture(GL_TEXTURE_2D, inputTextures[array]);
				glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texSize, texSize, GL_RGBA, GL_FLOAT, tmpfloat);
			}


			for (int iter = 0; iter < numKernelSteps; iter++) {
				//
				// assign accum texture (input pingpong) to kernel
				//
				glUniform1i(accumParam, 9);
				glActiveTexture(GL_TEXTURE9);
				glBindTexture(GL_TEXTURE_2D, pingtex[readTex]);

				for (int array=0; array < numArraysPerChunk ; array++) {
					glUniform1i(params[array], array);
					glActiveTexture(texunits[array]);
					glBindTexture(GL_TEXTURE_2D, inputTextures[array]);
				}

				// set rendering buffer
				glDrawBuffer(attachmentpoints[writeTex]);

				glEnable(GL_TEXTURE_2D);

				// start rendering
				glBegin(GL_QUADS);
				glTexCoord2f(0.0, 0.0);
				glVertex2f(0.0, 0.0);
				glTexCoord2f(texSize, 0.0);
				glVertex2f(texSize, 0.0);
				glTexCoord2f(texSize, texSize);
				glVertex2f(texSize, texSize);
				glTexCoord2f(0.0, texSize);
				glVertex2f(0.0, texSize);
				glEnd();
				glFlush();

				// swap pingpong input and output textures
				readTex = 1 - readTex;
				writeTex = 1 - writeTex;
			}
		}
	}

	glFinish();

	double end_compute = clock();
	//
	// Readback results (into double precision arrays).
	//
	printf("Reading back results\n");
	glFinish();

	double start_readback = clock();

	glFinish();

	double end_readback = clock();

	//
	// get computing result and comparison
	glReadBuffer(attachmentpoints[readTex]);
	glReadPixels(0, 0, texSize, texSize, GL_RGBA, GL_FLOAT, data);
	printf("Result after computing at the first four positions:\n");
	for (int i = 0; i < 4; i++) {
		printf("%d: %0.1f\n", i, data[i]);
	}
	double res = 0;
	for (int iter = 0; iter < numIterations; iter++) {
		for (int i = 0; i < numChunks; i++)
			for (int j = 0; j < numArraysPerChunk; j++)
				for (int k = 0; k < numKernelSteps; k++)
					res += i * 10.0 + j + 1;
	}
	printf("data[%d]:%.1f, correct value: %.1f\n", 0, data[0], res);

	//
	// print out timings
	//
	double time_compute = (end_compute - start_compute) / CLOCKS_PER_SEC;
	double time_readback = (end_readback - start_readback) / CLOCKS_PER_SEC;
	double data_download = (double) N * numChunks * numArraysPerChunk * numIterations * sizeof(float) / (1024.0 * 1024.0);
	double data_readback = (double) N * sizeof(float) / (1024.0 * 1024.0);
	printf("DOWNLOAD result from GPU side\n");
	printf("Total amount of data : %.2f MB\n", data_download);
	printf("Compute time         : %.2f sec\n", time_compute);
	printf("Throughput:          : %.2f MB/s\n", data_download / time_compute);
	printf("READBACK\n");
	printf("Total amount of data : %.2f MB\n", data_readback);
	printf("readback time         : %.2f sec\n", time_readback);
	printf("Throughput:          : %.2f MB/s\n", data_readback / time_readback);

	 //clean up
	for (int chunk = 0; chunk < numChunks; chunk++) {
		for (int array = 0; array < numArraysPerChunk; array++)
			delete[] chunks[chunk][array];
		delete[] chunks[chunk];
	}
	delete[] tmpfloat;
	delete[] data;
	delete[] chunks;
	glDeleteFramebuffersEXT(1, &fb);

	glDeleteTextures(numArraysPerChunk, inputTextures);
	glDeleteTextures(2, pingtex);

	return 0;
}


#endif


GLuint setupShader(const char **vertSrc, const char **fragSrc)
{
	GLuint	programObject;	// the program used to update
	GLuint	fragmentShader, vertexShader;

	programObject = glCreateProgram();

	if (fragSrc != NULL)
	{
		fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(fragmentShader, 1, fragSrc, NULL);
		glCompileShader(fragmentShader);
		glAttachShader(programObject, fragmentShader);
	}

	if (vertSrc != NULL)
	{
		vertexShader = glCreateShaderObjectARB(GL_VERTEX_SHADER);
		glShaderSource(vertexShader, 1, vertSrc, NULL);
		glCompileShader(vertexShader);
		glAttachShader(programObject, vertexShader);
	}

// Link the shader into a complete GLSL program.
	glLinkProgram(programObject);
	GLint progLinkSuccess;
	glGetShaderiv(programObject, GL_LINK_STATUS, &progLinkSuccess);
	if (!progLinkSuccess)
	{
		fprintf(stderr, "Shader could not be linked\n");
		exit(1);
	}
	return programObject;
}

void copy(float* in, float* out, int N) {
    for (int i=0; i<N; i++)
	out[i] = in[i];
}
void copy(float* in, void* out, int N) {
    float* bla = (float*)out;
    for (int i=0; i<N; i++)
	bla[i] = (float)in[i];
}
void copy(void* in, float* out, int N) {
    float* bla = (float*)in;
    for (int i=0; i<N; i++)
	out[i] = bla[i];
}


void CHECK_FRAMEBUFFER_STATUS()
{
	GLenum status;
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	printf("Framebuffer status %d\n", status);
	switch(status)
	{
		case GL_FRAMEBUFFER_COMPLETE_EXT:
			printf("Framebuffer complete\n");
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			printf("Framebuffer unsuported\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			printf("GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			printf("GL_FRAMEBUFFER_MISSING_ATTACHMENT_EXT\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			printf("GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			printf("GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			printf("GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			printf("GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT\n");
			break;
		default:
			printf("Framebuffer error\n");
	}
}
